/******************************************************************************
 * FIRESTARTER - A Processor Stress Test Utility 
 * Copyright (C) 2013 TU Dresden, Center for Information Services and High
 * Performance Computing
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Contact: daniel.hackenberg@tu-dresden.de
 *****************************************************************************/
 
#include "work.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <math.h>
#include <pthread.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <errno.h>

/*
 * assembler implementation of maximum power consumption measurement
 *
 */
int asm_work_avx_smt_compute(unsigned long long addrMem) __attribute__((noinline));
int asm_work_avx_smt_compute(unsigned long long addrMem)
{
/*
 * Works on Sandy-Bridge/Ivy-Bridge Processors (i.e. Model 42/58)
 */
 __asm__ __volatile__(
                "mfence;"
		"movabs $0x5555555555555555, %%r10;"
		"pinsrq $0, %%r10, %%xmm4;"
		"pinsrq $1, %%r10, %%xmm4;"
		"vinsertf128 $1, %%xmm4, %%ymm4, %%ymm4;"
		"movabs $0xAAAAAAAAAAAAAAAA, %%r10;"
		"pinsrq $0, %%r10, %%xmm5;"
		"pinsrq $1, %%r10, %%xmm5;"
		"vinsertf128 $1, %%xmm5, %%ymm5, %%ymm5;"

		//initialize avx-registers for boolean operations
		"vmovdqa %%ymm4, %%ymm0;"
		"vmovdqa %%ymm4, %%ymm2;"
		"vmovdqa %%ymm4, %%ymm6;" 
                "vmovdqa %%ymm5, %%ymm1;"
		"vmovdqa %%ymm5, %%ymm3;"
		"vmovdqa %%ymm5, %%ymm7;" 

		//initialize avx-registers for multiply operations
		"vmovdqa %%ymm4, %%ymm8;"
		"vmovdqa %%ymm4, %%ymm10;"
		"vmovdqa %%ymm4, %%ymm12;" 
		"vmovdqa %%ymm4, %%ymm14;" 
                "vmovdqa %%ymm5, %%ymm9;"
		"vmovdqa %%ymm5, %%ymm11;"
		"vmovdqa %%ymm5, %%ymm13;"
		"vmovdqa %%ymm5, %%ymm15;"                             

		//initialize GPRs for boolean operations
		"mov $0x0000000000000000, %%r8;"
		"mov $0x0000000000000000, %%r9;"
		"mov $0xFFFFFFFFFFFFFFFF, %%r10;"

		"movabs $4, %%rax;"
                "mfence;"
                ".align 64;"
                "_work_loop_avx_compute:"
		/***********************************************************************************************/ 
		"vmulpd %%ymm8,  %%ymm13, %%ymm8;  vxorpd %%ymm0, %%ymm5, %%ymm0; xor %%r10, %%r8; vsubpd %%ymm3, %%ymm5, %%ymm3;" 
		"vmulpd %%ymm9,  %%ymm12, %%ymm9;  vxorpd %%ymm1, %%ymm4, %%ymm1; xor %%r10, %%r9; vsubpd %%ymm2, %%ymm4, %%ymm2;" 
		"vmulpd %%ymm10, %%ymm13, %%ymm10; vxorpd %%ymm0, %%ymm4, %%ymm0; xor %%r10, %%r8; vaddpd %%ymm3, %%ymm5, %%ymm3;"
		"vmulpd %%ymm11, %%ymm12, %%ymm11; vxorpd %%ymm1, %%ymm5, %%ymm1; xor %%r10, %%r9; vaddpd %%ymm2, %%ymm4, %%ymm2;" 
                 
		"vmulpd %%ymm8,  %%ymm12, %%ymm8;  vxorpd %%ymm0, %%ymm5, %%ymm0; xor %%r10, %%r8; vsubpd %%ymm3, %%ymm5, %%ymm3;" 
		"vmulpd %%ymm9,  %%ymm13, %%ymm9;  vxorpd %%ymm1, %%ymm4, %%ymm1; xor %%r10, %%r9; vsubpd %%ymm2, %%ymm4, %%ymm2;"  
		"vmulpd %%ymm10, %%ymm12, %%ymm10; vxorpd %%ymm0, %%ymm4, %%ymm0; xor %%r10, %%r8; vaddpd %%ymm3, %%ymm5, %%ymm3;"  
		"vmulpd %%ymm11, %%ymm13, %%ymm11; vxorpd %%ymm1, %%ymm5, %%ymm1; xor %%r10, %%r9; vaddpd %%ymm2, %%ymm4, %%ymm2;"
                  
		"vmulpd %%ymm8,  %%ymm13, %%ymm8;  vxorpd %%ymm0, %%ymm5, %%ymm0; xor %%r10, %%r8; vsubpd %%ymm3, %%ymm5, %%ymm3;"   
		"vmulpd %%ymm9,  %%ymm12, %%ymm9;  vxorpd %%ymm1, %%ymm4, %%ymm1; xor %%r10, %%r9; vsubpd %%ymm2, %%ymm4, %%ymm2;" 
		"vmulpd %%ymm10, %%ymm13, %%ymm10; vxorpd %%ymm0, %%ymm4, %%ymm0; xor %%r10, %%r8; vaddpd %%ymm3, %%ymm5, %%ymm3;"
		"vmulpd %%ymm11, %%ymm12, %%ymm11; vxorpd %%ymm1, %%ymm5, %%ymm1; xor %%r10, %%r9; vaddpd %%ymm2, %%ymm4, %%ymm2;" 
                 
		"vmulpd %%ymm8,  %%ymm12, %%ymm8;  vxorpd %%ymm0, %%ymm5, %%ymm0; xor %%r10, %%r8; vsubpd %%ymm3, %%ymm5, %%ymm3;" 
		"vmulpd %%ymm9,  %%ymm13, %%ymm9;  vxorpd %%ymm1, %%ymm4, %%ymm1; xor %%r10, %%r9; vsubpd %%ymm2, %%ymm4, %%ymm2;"   
		"vmulpd %%ymm10, %%ymm12, %%ymm10; vxorpd %%ymm0, %%ymm4, %%ymm0; xor %%r10, %%r8; vaddpd %%ymm3, %%ymm5, %%ymm3;"  
		"vmulpd %%ymm11, %%ymm13, %%ymm11; vxorpd %%ymm1, %%ymm5, %%ymm1; xor %%r10, %%r9; vaddpd %%ymm2, %%ymm4, %%ymm2;"
		/*************************************************************************************************/               
		"sub $1, %%rax;"
		"jne _work_loop_avx_compute;"
		"movabs $4, %%rax;"
		/* Re-initialize AVX-Register for Multiplication */
		"vmovdqa %%ymm4, %%ymm8;"
		"vmovdqa %%ymm4, %%ymm10;"
		"vmovdqa %%ymm4, %%ymm12;" 
		"vmovdqa %%ymm4, %%ymm14;" 
                "vmovdqa %%ymm5, %%ymm9;"
		"vmovdqa %%ymm5, %%ymm11;"
		"vmovdqa %%ymm5, %%ymm13;"
		"vmovdqa %%ymm5, %%ymm15;"
		/*****************************/
		"sub $1, %%rbx;"
		"jmp _work_loop_avx_compute;"
                "mfence;"
                :
		: "a"(addrMem)
                : "%r8", "%r9", "%r10", "%r11", "%r12", "%r13", "%r14", "%r15", "%xmm0", "%xmm1", "%xmm2", "%xmm3", "%xmm4", "%xmm5", "%xmm6", "%xmm7", "%xmm8", "%xmm9", "%xmm10", "%xmm11", "%xmm12", "%xmm13", "%xmm14", "%xmm15"
		);

	return EXIT_SUCCESS;   
}


int asm_work_avx_smt_memory(unsigned long long addrMem) __attribute__((noinline));
int asm_work_avx_smt_memory(unsigned long long addrMem)
{  
/*
 * Works on Sandy-Bridge/Ivy-Bridge Processors (i.e. Model 42/58)
 */	
	__asm__ __volatile__(

                "mfence;"		

		/* Initialize SIMD-Register for L1-Writes */
		"vmovdqa (%%rax), %%ymm12;" //1111...
		"vmovdqa %%ymm12, %%ymm14;"
		"vmovdqa %%ymm12, %%ymm13;"
		"vmovdqa %%ymm12, %%ymm15;"

		"mov %%rax, %%rsi;"        //Save Bufferaddress in rsi + Address for L1-Read-Buffer in rax 
		"mov %%rax, %%rcx;"   
		"mov %%rax, %%r8;"     
		"mov %%rax, %%r10;"   
		"add $32768,   %%rdi;"     //Address for L1-Write-Buffer 
		"add $65536,   %%rcx;"     //Address for L2-Buffer in rcx	
		"add $327680,   %%r8;"     //Address for L3-Buffer in r8
		"add $2424832, %%r10;"     //Address for RAM-Buffer in r10

		"movabs $28,      %%rbx;"  //Counter for L1-Buffer  with 1152 Bytes per Loop  (32KB) (Read and Write)
		"movabs $682,     %%rdx;"  //Counter for L2-Buffer  with  384 Bytes per Loop  (256KB)
		"movabs $16384,    %%r9;"  //Counter for L3-Buffer  with  128 Bytes per Loop  (2MB per physical Core)
		"movabs $4194304, %%r11;"  //Counter for RAM-Buffer with   64 Bytes per Loop  (256MB per Core)

                "mfence;"
                ".align 64;"
                "_work_loop_avx_smt_memory:"
		/**************************************************************/
		"prefetcht0 (%%r10);" 

		"vmovdqa    (%%r8),  %%ymm8;"
		"vmovdqa    (%%rcx), %%ymm4; vmovdqa %%ymm12, (%%rdi);"
		"vmovdqa    (%%rax), %%ymm0; vmovdqa  64(%%rax), %%ymm1; vmovdqa 128(%%rax), %%ymm2;"
		"vmovdqa  64(%%rcx), %%ymm5; vmovdqa %%ymm13, 64(%%rdi);"
		"vmovdqa 192(%%rax), %%ymm0; vmovdqa 256(%%rax), %%ymm1; vmovdqa 320(%%rax), %%ymm2;"
		"vmovdqa 128(%%rcx), %%ymm6; vmovdqa %%ymm14, 128(%%rdi);"
		"vmovdqa 384(%%rax), %%ymm0; vmovdqa 448(%%rax), %%ymm1; vmovdqa 512(%%rax), %%ymm2;"
		
		"vmovdqa   64(%%r8), %%ymm9;"
		"vmovdqa 192(%%rcx), %%ymm4; vmovdqa %%ymm12, 192(%%rdi);"
		"vmovdqa 576(%%rax), %%ymm0; vmovdqa 640(%%rax), %%ymm1; vmovdqa 704(%%rax), %%ymm2;"
		"vmovdqa 256(%%rcx), %%ymm5; vmovdqa %%ymm13, 256(%%rdi);"
		"vmovdqa 768(%%rax), %%ymm0; vmovdqa 832(%%rax), %%ymm1; vmovdqa 896(%%rax), %%ymm2;"
		"vmovdqa 320(%%rcx), %%ymm6; vmovdqa %%ymm14, 320(%%rdi);"
		"vmovdqa 960(%%rax), %%ymm0; vmovdqa 1024(%%rax), %%ymm1; vmovdqa 1088(%%rax), %%ymm2;"			
		/**************************************************************/
		//RAM		
		"add $64, %%r10;"
		"sub $1, %%r11;"
		"jne _work_L1_avx_smt_memory;"
		"mov %%rsi, %%r10;"
		"add $2424832, %%r10;"
		"movabs $4194304, %%r11;"
		//L1-Cache 
		"_work_L1_avx_smt_memory:"
		"add $1152, %%rax;" //L1-Reads
		"add $1152, %%rdi;" //L1-Writes
		"sub $1, %%rbx;"
		"jne _work_L2_avx_smt_memory;"
		"movabs $28, %%rbx;" 
		"mov %%rsi, %%rax;"
		"mov %%rsi, %%rdi;"  //L1-Writes
		"add $32768, %%rdi;" //L1-Writes
		//L2-Cache
		"_work_L2_avx_smt_memory:"
		"add $384, %%rcx;"
		"sub $1, %%rdx;"
		"jne _work_L3_avx_smt_memory;"
		"movabs $682, %%rdx;"
		"mov %%rsi, %%rcx;"
		"add $65536, %%rcx;"
		//L3-Cache
		"_work_L3_avx_smt_memory:"
		"add $128, %%r8;"
		"sub $1, %%r9;"
		"jne _work_loop_avx_smt_memory;"	
		"movabs $16384, %%r9;"
		"mov %%rsi, %%r8;"
		"add $327680, %%r8;"
		"jmp _work_loop_avx_smt_memory;"
                "mfence;"
                : 
		: "a"(addrMem)
                : "%r8", "%r9", "%r10", "%r11", "%r12", "%r13", "%r14", "%r15", "%xmm0", "%xmm1", "%xmm2", "%xmm3", "%xmm4", "%xmm5", "%xmm6", "%xmm7", "%xmm8", "%xmm9", "%xmm10", "%xmm11", "%xmm12", "%xmm13", "%xmm14", "%xmm15"
		);

	return EXIT_SUCCESS;  	
}

/*
 * function that does the measurement
 */
inline void _work( unsigned long long memsize,int function,volatile mydata_t* data)
{
  int i,t;
 
  /* aligned addresses */
  unsigned long long addrMem; 	
  addrMem= (unsigned long long)(data->bufferMem);
  data->ack=0;
  int tmp;

 // printf("starting measurment %i accesses in %i Bytes of memory\n",accesses,memsize);
   t=data->num_threads-1;
   data->running_threads=t+1;

    //init threaddata
    if (t)
    {
     for (i=0;i<t;i++)
     {
      data->threaddata[i+1].memsize=memsize/(t+1);
      }
    }    

    //tell other threads to touch memory 
     if (t)
     {
       for (i=0;i<t;i++)
       {
        data->thread_comm[i+1]=THREAD_USE_MEMORY;
        while (!data->ack);
        data->ack=0;
       }
     }
         
     //wait for other THREAD touching the memory
     if (t)
     {
       for (i=0;i<t;i++)
       {
         data->thread_comm[i+1]=THREAD_WAIT;       
         while (!data->ack);
         data->ack=0;
       }
     }
    
     //tell other threads to start
     if (t)
     {
       for (i=0;i<t;i++)
       {
         data->thread_comm[i+1]=THREAD_WORK;
         while (!data->ack);
         data->ack=0;
       }
     }      
      
      /* call ASM implementation */
     switch(function)
     {
  	case 0:
		tmp=asm_work_sse(addrMem);
		break;
	case 1:
		tmp=asm_work_sse_smt(addrMem);
		break;
	case 2:
		tmp=asm_work_avx(addrMem);
		break;
	case 3:
        	tmp=asm_work_avx_smt_compute(addrMem);
		break;
	case 4:
		tmp=asm_work_model45_avx(addrMem);
		break;
	case 5:
		tmp=asm_work_model45_avx_smt(addrMem);
		break;
	default:
         	fprintf(stderr,"Error: unknown function %i\n",function);
         	exit(1);
     }   
	if (tmp != EXIT_SUCCESS){
		fprintf(stderr,"Error in function %i\n",function);
         	exit(1);
	} 
} 

/*
 * loop for additional worker threads
 * communicating with master thread using shared variables
 */
 void *thread(void *threaddata)
{
  int id= ((threaddata_t *) threaddata)->thread_id;
  volatile mydata_t* global_data = ((threaddata_t *) threaddata)->data; //communication
  threaddata_t* mydata = (threaddata_t*)threaddata;
  unsigned short half_num_threads=(unsigned short)global_data->num_threads / 2;

  unsigned int i,j;
  double tmp=(double)0;
  unsigned long long tmp2,old=THREAD_STOP;  
 
  do
  {
   old=global_data->thread_comm[id];
  }
  while (old!=THREAD_INIT);
  global_data->ack=id;

  cpu_set(((threaddata_t *) threaddata)->cpu_id); 

  if(mydata->buffersizeMem)
  {
	 
	switch((unsigned char)mydata->FUNCTION)
	{
	case 0:
	case 1:
        case 2:
	case 3:
        case 4:
	case 5:
		mydata->bufferMem   = _mm_malloc(mydata->buffersizeMem,mydata->alignment);
		break;
	default:
		fprintf( stderr, "Error: initialization failed: unknown mode:%i \n",mydata->FUNCTION );      
    		exit( 1 );
	}
  
	switch((unsigned char)mydata->FUNCTION)
	{
	case 0:
	case 1:
	case 2:
	case 4:
	case 5:
		mydata->addrMem=(unsigned long long)(mydata->bufferMem);
		break; 	
	case 3:	//AVX-SMT Model 42,58
		// Fill Cache-/Memory-buffer   
   		tmp2=32*mydata->BURST_LENGTH*sizeof(double);
		// Fill RAM
       	 	for (i=0;i<=mydata->buffersizeMem-tmp2;i+=tmp2)
        	{
			for(j=0;j<4*mydata->BURST_LENGTH;j++)
 			  *((unsigned long long*)((unsigned long long)mydata->bufferMem+i+j*sizeof(unsigned long long)))=18446744073709551615llu; //1111...
        		  //*((double*)((unsigned long long)mydata->bufferMem+i+j*sizeof(double)))=global_data->FP_INIT[0];
      			for(j=4*mydata->BURST_LENGTH;j<8*mydata->BURST_LENGTH;j++)
        		  *((double*)((unsigned long long)mydata->bufferMem+i+j*sizeof(double)))=-1.0*global_data->FP_INIT[1];
      			for(j=8*mydata->BURST_LENGTH;j<12*mydata->BURST_LENGTH;j++)
       			  *((double*)((unsigned long long)mydata->bufferMem+i+j*sizeof(double)))=global_data->FP_INIT[0];
    			for(j=12*mydata->BURST_LENGTH;j<16*mydata->BURST_LENGTH;j++)
        		  *((double*)((unsigned long long)mydata->bufferMem+i+j*sizeof(double)))=-1.0*global_data->FP_INIT[1];
      			for(j=16*mydata->BURST_LENGTH;j<20*mydata->BURST_LENGTH;j++)
        		  *((double*)((unsigned long long)mydata->bufferMem+i+j*sizeof(double)))=-1.0*global_data->FP_INIT[0];
      			for(j=20*mydata->BURST_LENGTH;j<24*mydata->BURST_LENGTH;j++)
        		  *((double*)((unsigned long long)mydata->bufferMem+i+j*sizeof(double)))=global_data->FP_INIT[1];
      			for(j=24*mydata->BURST_LENGTH;j<28*mydata->BURST_LENGTH;j++)
        		  *((double*)((unsigned long long)mydata->bufferMem+i+j*sizeof(double)))=-1.0*global_data->FP_INIT[0];
      			for(j=28*mydata->BURST_LENGTH;j<32*mydata->BURST_LENGTH;j++)
        		  *((double*)((unsigned long long)mydata->bufferMem+i+j*sizeof(double)))=global_data->FP_INIT[1];
   		}
		mydata->addrMem=(unsigned long long)(mydata->bufferMem);
		break;
   default:
   	fprintf( stderr, "Error: initialization failed: unknown mode:%i \n",mydata->FUNCTION );      
    	exit( 1 );
   }
  }  

  while(1)
  {
     //printf("Thread %i: comm= %i\n",id+1,data->thread_comm[id]);
     switch (global_data->thread_comm[id]){
       case THREAD_USE_MEMORY: 
         if (old!=THREAD_USE_MEMORY)
         {
           if ((old==THREAD_WAIT)||(old==THREAD_WORK)||(old==THREAD_INIT))
           {

           }
           old=THREAD_USE_MEMORY;
           global_data->ack=id;

           global_data->done=id;
         }
         else 
         {
           tmp=100;while(tmp>0) tmp--; 
         }        
         break;
       case THREAD_WAIT: // waiting
          if (old!=THREAD_WAIT)
          {
            old=THREAD_WAIT;
            global_data->ack=id;

          }
          tmp=100;while(tmp) tmp--;
          break;
       case THREAD_WORK:
          if (old!=THREAD_WORK) 
          {
           if ((old==THREAD_WAIT)||(old==THREAD_INIT))
           {

           }
            old=THREAD_WORK;
            global_data->ack=id;

	
        /* call ASM implementation */
        switch ((unsigned char)mydata->FUNCTION)
        {
	case 0:
		tmp=asm_work_sse(mydata->addrMem);
		break;
	case 1:
		tmp=asm_work_sse_smt(mydata->addrMem);
		break;
	case 2:
		tmp=asm_work_avx(mydata->addrMem);
		break;
	case 3:
		if (id < (int)half_num_threads){
		tmp=asm_work_avx_smt_compute(mydata->addrMem);
		}
		else{
		tmp=asm_work_avx_smt_memory(mydata->addrMem);
		} 
		break;
	case 4:
		tmp=asm_work_model45_avx(mydata->addrMem);
		break;
	case 5:
		tmp=asm_work_model45_avx_smt(mydata->addrMem);
		break;
        default:
                fprintf(stderr,"Error: unknown function %i\n",mydata->FUNCTION);
                pthread_exit(NULL);
                break;
	}
            
          global_data->done=id;

          }
          else 
          {
            tmp=100;while(tmp>0) tmp--;
          }  
          break;
       case THREAD_INIT: // used for parallel initialisation only
          tmp=100;while(tmp) tmp--;
          break;
       case THREAD_STOP: // exit
       default:
         pthread_exit(NULL);
    }
  }
}

