/******************************************************************************
 * FIRESTARTER - A Processor Stress Test Utility 
 * Copyright (C) 2013 TU Dresden, Center for Information Services and High
 * Performance Computing
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Contact: daniel.hackenberg@tu-dresden.de
 *****************************************************************************/

#ifndef __WORK_H
#define __WORK_H

#include <pthread.h>
#include <mm_malloc.h>
#include "arch.h"

#define THREAD_USE_MEMORY  1
#define THREAD_WAIT        2
#define THREAD_FLUSH       3
#define THREAD_WORK        4
#define THREAD_INIT        5
#define THREAD_STOP        6

/** The data structure that holds all the global data.
 */
typedef struct mydata
{
   char* bufferMem;					
   pthread_t *threads;					
   struct threaddata *threaddata;			
   cpu_info_t *cpuinfo;					
   long long *INT_INIT;
   double *FP_INIT;				
   unsigned short num_threads;				
   int *thread_comm;					
   volatile unsigned int running_threads;  
   volatile unsigned short ack;				
   volatile unsigned short done;			
   int *threads_per_package;
   unsigned int buffersize[3];
   unsigned long long rambuffersize;
} mydata_t;

/* data needed by each thread */
typedef struct threaddata
{
   volatile mydata_t *data;				
   char* bufferMem;					
   cpu_info_t *cpuinfo;					
   unsigned long long addrMem;		
   unsigned int thread_id;
   unsigned int memsize;     
   unsigned int accesses;				
   unsigned long long buffersizeMem;     
   unsigned int buffersize[3];
   unsigned int alignment;      
   unsigned int cpu_id;          				
   unsigned char FUNCTION;
   unsigned char BURST_LENGTH;
   int package;
} threaddata_t;

/*
 * function that does the measurement
 */
extern void _work(unsigned long long memsize,int function,volatile mydata_t* data);

/*
 * loop executed by all threads, except the master thread
 */
extern void *thread(void *threaddata);

/*
 * prototypes 
 */
int asm_work_sse(unsigned long long addrMem) __attribute__((noinline));
int asm_work_sse(unsigned long long addrMem);

int asm_work_sse_smt(unsigned long long addrMem) __attribute__((noinline));
int asm_work_sse_smt(unsigned long long addrMem);

int asm_work_avx(unsigned long long addrMem) __attribute__((noinline));
int asm_work_avx(unsigned long long addrMem);

int asm_work_model45_avx_smt(unsigned long long addrMem) __attribute__((noinline));
int asm_work_model45_avx_smt(unsigned long long addrMem);

int asm_work_model45_avx(unsigned long long addrMem) __attribute__((noinline));
int asm_work_model45_avx(unsigned long long addrMem);
#endif
